/*
 *     Refactor.java  Nov 23, 2005
 *
 *     Katsuhisa Maruyama (maru@jtool.org)
 */

package org.jtool.jrbx.gui;
import org.jtool.platform.gui.*;
import org.jtool.platform.j2xml.*;
import org.jtool.platform.event.*;
import org.jtool.platform.xml.JXConstants;
import org.jtool.platform.xml.dom.*;
import org.jtool.jrbx.refactor.util.CodeSelection;
import org.jtool.jrbx.refactor.util.UndoChecker;
import org.jtool.jrbx.refactor.dialog.MessageDialog;
import org.jtool.jrbx.refactor.classes.ClassRefactoring;
import org.jtool.jrbx.refactor.methods.MethodRefactoring;
import org.jtool.jrbx.refactor.fields.FieldRefactoring;
import org.jtool.jrbx.refactor.statements.StatementRefactoring;
import org.jtool.jrbx.refactor.variables.VariableRefactoring;
import org.jtool.jrbx.refactor.miscellaneous.MiscellaneousRefactoring;
import org.jtool.jrbx.refactor.RefactoringCommand;
import javax.swing.JOptionPane;
import javax.swing.JFrame;
import java.util.*;
import java.io.*;
import org.w3c.dom.*;

public class Refactor extends EventSource {
    private CodesPane codesPane;
    private RefactoringCommand commander;
    private MainFrame frame;
    private int refactoringId = 1;
    private CodeSelection selection;

    static final String CLASS = "Class";
    static final String METHOD = "Method";
    static final String FIELD = "Field";
    static final String VARIABLE = "Variable";
    static final String STATEMENT = "Statement";
    static final String MISCELLANEOUS = "Miscellaneous";

    public Refactor(MainFrame frame) {
        this.frame = frame;
    }

    public void setCodesPane(CodesPane pane) {
        codesPane = pane;
    }

    public JFrame getFrame() {
        return frame;
    }

    public void setSelection(CodePane codePane, int start, int end) {
        parseFilesOnEdit();
        
        Document doc = codePane.getXMLDocument();
        selection = new CodeSelection(doc, start, end);
        selection.find();
    }

    private void parseFilesOnEdit() {
        FileManager fileManager = frame.getFileManager();
        fileManager.parseFilesOnEdit();
    }

    public boolean isClassRefactoring() {
        return ClassRefactoring.isEnabled(selection);
    }

    public boolean isMethodRefactoring() {
        return MethodRefactoring.isEnabled(selection);
    }

    public boolean isFieldRefactoring() {
        return FieldRefactoring.isEnabled(selection);
    }

    public boolean isVariableRefactoring() {
        return VariableRefactoring.isEnabled(selection);
    }

    public boolean isStatementRefactoring() {
        return StatementRefactoring.isEnabled(selection);
    }

    public boolean isSwitchRefactoring() {
        return StatementRefactoring.isEnabledSwitch(selection);        
    }

    public boolean isIfRefactoring() {
        return StatementRefactoring.isEnabledIf(selection);
    }

    public boolean isWhileRefactoring() {
        return StatementRefactoring.isEnabledWhile(selection);
    }

    public boolean isDoRefactoring() {
        return StatementRefactoring.isEnabledDo(selection);
    }

    public boolean isForRefactoring() {
        return StatementRefactoring.isEnabledFor(selection);
    }

    public boolean isExtractMethodRefactoring() {
        return MiscellaneousRefactoring.isEnabled(selection);
    }

    public void execute(String category, String command) {
        if (category.equals(CLASS)) {
            commander = ClassRefactoring.create(command);

        } else if (category.equals(METHOD)) {
            commander = MethodRefactoring.create(command);

        } else if (category.equals(FIELD)) {
            commander = FieldRefactoring.create(command);

        } else if (category.equals(VARIABLE)) {
            commander = VariableRefactoring.create(command);

        } else if (category.equals(STATEMENT)) {
            commander = StatementRefactoring.create(command);

        } else if (category.equals(MISCELLANEOUS)) {
            commander = MiscellaneousRefactoring.create(command);

        } else {
            return;
        }

        CodePane codePane = codesPane.getCurrentCodePane();
        commander.setSource(codePane.getXMLDocument(), selection);
        commander.setRefactor(this);
        commander.execute();

        String log = commander.getRefactoringLog();
        if (log != null) {
            updateTextPane(log, codePane.getFileName(), commander.getUndoInformation());

            String refactoringNum = "#" + String.valueOf(refactoringId);
            fire(new LogEvent(this, refactoringNum + ": " + log));
            refactoringId++;
        }

        codesPane.setSelectedComponent(codePane);
        codesPane.updateUndoState();
    }

    private void updateTextPane(String command, String curPathName, String undoInfo) {
        FileManager fileManager = frame.getFileManager();
        ArrayList fileNames = getFileNameList(commander.getChangedXMLTexts());

        Iterator it = commander.getChangedXMLTexts().iterator();
        while (it.hasNext()) {
            JavaXMLText text = (JavaXMLText)it.next();
            String pathName = text.getName();

            CodePane codePane = codesPane.getCodePane(pathName);
            if (codePane == null) {
                File file = new File(pathName);
                if (file.exists()) {
                    codePane = fileManager.openFile(pathName);
                } else {
                    codePane = fileManager.newFile(pathName);
                }
                codePane.getUndoManager().discardAllEdits();
            }

            String key = getKey(command, codePane.getFileName(), curPathName);
            RefactoringRecord record = new RefactoringRecord(key, codePane.getFileName(), undoInfo, fileNames);

            RefactoringUndoManager undoManager = (RefactoringUndoManager)codePane.getUndoManager();
            undoManager.beginCompoundEdit();
            codePane.setText(text.getCode());
            undoManager.endCompoundEdit(record);
            undoManager.memorize();
            
            codePane.setChange(true);
            codePane.setCaretPosition(0);
            codePane.validate();
        }
    }

    private ArrayList getFileNameList(ArrayList texts) {
        ArrayList fileNames = new ArrayList();

        Iterator it = texts.iterator();
        while (it.hasNext()) {
            JavaXMLText text = (JavaXMLText)it.next();
            fileNames.add(text.getName());
        }
        return fileNames;
    }

    private String getKey(String command, String pathName, String curPathName) {
        if (pathName.equals(curPathName)) {
            return command + "#" + String.valueOf(refactoringId) + ":src";
        } else {
            return command + "#" + String.valueOf(refactoringId) + ":dst";
        }
    }

    public String canUndo(CodePane codePane) {
        RefactoringUndoManager undoManager = (RefactoringUndoManager)codePane.getUndoManager();
        RefactoringRecord record = undoManager.getLastRefactoringCommand();
        if (record != null && canUndo(record)) {
            return record.getCommandWithoutId();
        }

        return null;
    }

    private boolean canUndo(RefactoringRecord record) {
        Iterator it = record.getChangedFiles().iterator();
        while (it.hasNext()) {
            String name = (String)it.next();

            CodePane codePane = codesPane.getCodePane(name);
            if (codePane == null) {
                return false;
            }
            if (!canUndo(codePane, record)) {
                return false;
            }
        }
        return true;
    }

    private boolean canUndo(CodePane codePane, RefactoringRecord record) {
        RefactoringUndoManager undoManager = (RefactoringUndoManager)codePane.getUndoManager();
        return undoManager.isLastRefactoringCommand(record.getCommand());
    }

    public void undo(CodePane codePane) {
        RefactoringUndoManager undoManager = (RefactoringUndoManager)codePane.getUndoManager();
        RefactoringRecord record = undoManager.getLastRefactoringCommand();

        ArrayList changedFiles = record.getChangedFiles();
        String command = record.getCommandWithoutId();

        if (!confirmDiscardEditoralChanges(command, changedFiles)) {
            return;
        }

        if (!confirmUndo(command, changedFiles)) {
            return;
        }

        parseFilesOnEdit();
        if (UndoChecker.check(codePane.getFileName(), record.getUndoInformation())) {
            MessageDialog messageDialog = new MessageDialog(frame);
            messageDialog.show("Undo " + command + " might change the behavior of source code.");
            return;
        }

        Iterator it = changedFiles.iterator();
        while (it.hasNext()) {
            String name = (String)it.next();
            CodePane pane = codesPane.getCodePane(name);
            RefactoringUndoManager um = (RefactoringUndoManager)pane.getUndoManager();

            um.undoRefactoring();
            um.memorize();
            codePane.setCaretPosition(0);
            codePane.setChange(true);
        }
        
        codesPane.updateUndoState();
        fire(new LogEvent(this, "Undo: " + record.getSimpleLog()));
    }

    private boolean confirmDiscardEditoralChanges(String command, ArrayList changedFiles) {
        ArrayList files = new ArrayList();
        Iterator it = changedFiles.iterator();
        while (it.hasNext()) {
            String name = (String)it.next();

            CodePane codePane = codesPane.getCodePane(name);
            RefactoringUndoManager undoManager = (RefactoringUndoManager)codePane.getUndoManager();
            if (!undoManager.isLastChangeByRefactoring()) {
                files.add(name);
            }
        }

        if (files.size() == 0) {
            return true;
        }

        int result = JOptionPane.showConfirmDialog(frame, 
          "The editoral changes were made to the following file(s):\n" + getPrettyFileNameList(files)
          + "Do you want to discard the changes by undoing " + command +"?",
          "Select an Option", JOptionPane.YES_NO_OPTION);
        if (result == JOptionPane.YES_OPTION) {
            return true;
        }
        return false;
    }

    private boolean confirmUndo(String command, ArrayList files) {
        int result = JOptionPane.showConfirmDialog(frame, "Undoing " + command + ":\n"
          + "The changes were made to the following file(s):\n" + getPrettyFileNameList(files)
          + "Do you want to reverse the changes?", "Select an Option", JOptionPane.YES_NO_OPTION);
        if (result == JOptionPane.YES_OPTION) {
            return true;
        }
        return false;
    }

    private String getPrettyFileNameList(List files) {
        StringBuffer buf = new StringBuffer();

        Iterator it = files.iterator();
        while (it.hasNext()) {
            String name = (String)it.next();
            buf.append("\"");
            buf.append(name);
            buf.append("\"\n");
        }
        return buf.toString();
    }
}
