/*
 *     MoveMethodTransformer.java  Aug 5, 2004
 *
 *     Katsuhisa Maruyama (maru@jtool.org)
 */

package org.jtool.jrbx.refactor.methods;
import org.jtool.jrbx.refactor.util.*;
import org.jtool.jrbx.refactor.query.*;
import org.jtool.platform.xml.wrapper.*;
import org.jtool.platform.xml.JXConstants;
import java.util.*;
import org.w3c.dom.*;

public class MoveMethodTransformer extends RefactoringTransformer {
    private JavaMethod jmethod;
    private String refName;
    private boolean deleteParam;
    private boolean insertParam;
    private int paramNum;

    private MoveMethodTransformer() {
    }

    public MoveMethodTransformer(JavaMethod jmethod, String refName, boolean deleteParam, boolean insertParam) {
        this.jmethod = jmethod;
        this.refName = refName;
        this.deleteParam = deleteParam;
        this.insertParam = insertParam;
    }

    public void changeSrc() {
        delete(jmethod);

        paramNum = jmethod.getNumParameter();
        if (deleteParam) {
            paramNum = paramNum - 1;
        }

        if (refName != null && !jmethod.isStatic()) {
            insertDelegationMethod();
        }
    }

    public void changeDst(JavaClass dclass) {
        Node node = insert(jmethod, dclass);
        JavaMethod jm = new JavaMethod(node);

        insertImport(jmethod.getJavaClass(), dclass);
        
        if (refName != null) {
            if (deleteParam) {
                deleteRefParameter(jm);
            }
            eliminatePrefix(jm);
        }

        if (insertParam) {
            String type = jmethod.getJavaClass().getName();
            String name = type.toLowerCase();
            while (alreadyUsed(jm, name)) {
                name = "_" + name;
            }

            insertParameter(jm, type, name);
            replaceThisWithVariableName(jm, name);
            addPrefixName(jm, type, name);

            String fqn = jmethod.getJavaClass().getFqn();
            insertImport(dclass, jmethod.getJavaClass()); 
        }

        if (jmethod.isStatic()) {
            eliminatePrefixType(jm);
            addPrefixName(jm, jmethod.getJavaClass().getName(), "");
        }

        addPrefixType(jm, jmethod.getJavaClass().getName());
        insertImport(dclass, jm);
    }

    private void insertDelegationMethod() {
        MethodCode mcode = new MethodCode(jmethod);

        StringBuffer buf = new StringBuffer();
        buf.append("    ");
        buf.append(mcode.getCode());
        buf.append(" {\n");

        buf.append("        ");
        if (!jmethod.isVoid()) {
            buf.append("return ");
        }

        buf.append(refName);
        buf.append(".");
        buf.append(jmethod.getName());
        buf.append("(");
        if (insertParam) {
            if (paramNum == 0) {
                buf.append("this");
            } else {
                buf.append("this, ");
            }
        }
        if (deleteParam) {
            buf.append(mcode.getParameterNamesExcept(refName));
        } else {
            buf.append(mcode.getParameterNames());
        }
        buf.append(");\n");
        buf.append("    }");

        insertCodeBefore(buf.toString(), jmethod.getElem());
    }

    private JavaParameter deleteRefParameter(JavaMethod jm) {
        Iterator it = jm.getParameterNodes().iterator();
        while (it.hasNext()) {
            JavaParameter jp = new JavaParameter((Node)it.next());

            if (jp.getName().equals(refName)) {
                delete(jp);
                return jp;
            }
        }
        return null;
    }

    private void eliminatePrefix(JavaMethod jm) {
        Iterator it = jm.getUseVariableList().iterator();
        while (it.hasNext()) {
            JavaVariable jv = (JavaVariable)it.next();

            if (jv.getName().equals(refName) &&
                jv.getTypeFqn().equals(jm.getJavaClass().getFqn())) {

                if (jv.isDotChild()) {
                    Element prefix = (Element)jv.getDotPrefix();
                    if (prefix == null) {
                        delete(jv);

                    } else {
                        if (prefix.getNodeName().equals(JXConstants.ExpressionElem)) {
                            JavaExpr jr = new JavaExpr(prefix);
                            if (jr.isThis()) {
                                rename("this", jv);
                            } else {
                                delete(jv);
                            }
                        }
                    }

                } else {
                    rename("this", jv);
                }
            }
        }
    }

    private void insertParameter(JavaMethod jm, String type, String name) {
        Node node = jm.getFirstChildNode(jm.getElem(), JXConstants.OperatorElem);
        if (node != null) {
            if (paramNum == 0) {
                insertCodeAfter(type + " " + name, node);
            } else {
                insertCodeAfter(type + " " + name + ", ", node);
            }
        }
    }

    private void insertImport(JavaClass dclass, JavaClass jc) {
        if (QueryImport.needsImport(dclass, jc)) {
            insertImportCode(jc.getFqn(), dclass.getJavaFile());
        }
    }

    private void insertImport(JavaClass dclass, JavaMethod jm) {
        ArrayList imports = new ArrayList();
        JavaClassList cl = jm.getAllInnerClassesAndInterfaces();

        Iterator it = jm.getUsedTypes().iterator();
        while (it.hasNext()) {
            JavaClass jc = (JavaClass)it.next();
            if (!jc.equals(jm.getJavaClass()) && !cl.contains(jc) &&
                !imports.contains(jc.getFqn())) {
                insertImport(dclass, jc);
                imports.add(jc.getFqn());
            }
        }
    }

    private boolean alreadyUsed(JavaMethod jm, String name) {
        Iterator it = jm.getVariableList().iterator();
        while (it.hasNext()) {
            JavaVariable jv = (JavaVariable)it.next();
            if (jv.getName().equals(name)) {
                return true;
            }
        }
        return false;
    }

    private void replaceThisWithVariableName(JavaMethod jm, String name) {
        Iterator it = QueryMethod.getJavaExprThis(jm).iterator();
        while (it.hasNext()) {
            JavaExpr jr = (JavaExpr)it.next();
            delete(jr);

            Element opelem = (Element)jr.getNextSibling();
            if (opelem != null && opelem.getNodeName().equals(JXConstants.OperatorElem) &&
                opelem.getFirstChild().getNodeValue().equals(".")) {
                Element elem = getNextSibling(opelem);

                JavaExpr jr2 = new JavaExpr(elem);
                if (jr2.isVarRef()) {
                    JavaVariable jv = new JavaVariable(jr2.getElem());
                    if (jv.getName().equals(refName) &&
                        jv.getTypeFqn().equals(jm.getJavaClass().getFqn())) {
                        delete(opelem);
                        continue;
                    }
                }
            }
            insertCodeAfter(name, jr.getElem());
        }
    }

    private void addPrefixName(JavaMethod jm, String type, String name) {
        Iterator it = QueryMethod.getJavaExprMethodCall(jm).iterator();
        while (it.hasNext()) {
            JavaExpr jr = (JavaExpr)it.next();

            if (jr.getDotPrefix() == null) {
                JavaCall ja = new JavaCall(jr.getElem());
                if (ja.getCalledJavaMethod().isStatic()) {
                    insertCodeBefore(type + ".", ja.getElem());
                } else {
                    insertCodeBefore(name + ".", ja.getElem());
                }
            }
        }
        
        it = QueryMethod.getJavaExprVarRef(jm).iterator();
        while (it.hasNext()) {
            JavaExpr jr = (JavaExpr)it.next();

            if (jr.getDotPrefix() == null) {
                JavaVariable jv = new JavaVariable(jr.getElem());
                if (jv.isField() && !jv.getName().equals(refName)) {
                    if (jv.getUsedJavaField().isStatic()) {
                        insertCodeBefore(type + ".", jv.getElem());
                    } else {
                        insertCodeBefore(name + ".", jv.getElem());
                    }
                }
            }
        }
    }

    private void eliminatePrefixType(JavaMethod jm) {
        Iterator it = jm.getTypeNodes().iterator();
        while (it.hasNext()) {
            Element elem = (Element)it.next();

            JavaType jt = new JavaType(elem);
            if (jt.isDotChild() && jt.getName().equals(refName)) {
                delete(jt);
            }
        }
    }

    private void addPrefixType(JavaMethod jm, String type) {
        String pname = jmethod.getJavaFile().getPackageName();

        Iterator it = jm.getTypeNodes().iterator();
        while (it.hasNext()) {
            Element elem = (Element)it.next();
            JavaType jt = new JavaType(elem);

            if (jt.getName().indexOf(".") == -1) {
                addPrefixType(jt, type);

            } else {
                String tname = jt.getName();
                if (tname.startsWith(pname + ".")) {
                    tname = tname.substring(pname.length() + 1);
                }
                if (!tname.startsWith(type + ".")) {
                    addPrefixType(jt, type);
                }
            }
        }
    }

    private void addPrefixType(JavaType jt, String type) {
        JavaClassList cl = jmethod.getAllInnerClassesAndInterfaces();

        Iterator it = jmethod.getJavaClass().getAllInnerClassesAndInterfaces().iterator();
        while (it.hasNext()) {
            JavaClass jc = (JavaClass)it.next();

            if (jc.getFqn().equals(jt.getPrefixFqn()) && !cl.contains(jc)) {
                insertCodeBefore(type + ".", jt.getElem());
            }
        }
    }

    public void changeRef(JavaFile rfile) {
        Iterator it = QueryFile.getJavaCallsUsingTheMethod(rfile, jmethod).iterator();
        while (it.hasNext()) {
            JavaCall ja = (JavaCall)it.next();

            Element elem = (Element)ja.getDotPrefix();
            if (elem != null && elem.getNodeName().equals(JXConstants.TypeElem)) {
                rename(refName, new JavaType(elem));
            } else {
                insertCodeBefore(refName + ".", ja.getElem());
            }
        }
    }
}
