/*
 *     ExtractMethod.java  Aug 29, 2004
 *
 *     Katsuhisa Maruyama (maru@jtool.org)
 */

package org.jtool.jrbx.refactor.miscellaneous;
import org.jtool.jrbx.refactor.*;
import org.jtool.jrbx.refactor.util.*;
import org.jtool.jrbx.refactor.query.*;
import org.jtool.jrbx.refactor.dialog.*;
import org.jtool.platform.xml.JXConstants;
import org.jtool.platform.xml.dom.*;
import org.jtool.platform.xml.wrapper.*;
import org.jtool.platform.xml.graph.*;
import org.jtool.platform.xml.graph.cfg.*;
import org.jtool.platform.xml.graph.pdg.*;
import org.w3c.dom.*;
import java.util.*;
import javax.swing.JFrame;

public class ExtractMethod extends MiscellaneousRefactoring {
    private String newName = null;
    private JavaVariable rjv = null;
    private ExtractMethodTransformer transformer;
    private String modifier;
    private MethodCode mcode;
    private CFG cfg;
    private PDG pdg;
    private GraphCompSet innerCFGNodes;
    private GraphCompSet outerCFGNodes;
    private GraphCompSet innerPDGNodes;

    public ExtractMethod() {
        super();
    }

    public void setUpWithOptions(JFrame f, String newName, String modifier) {
        frame = f;
        this.newName = newName;
        this.modifier = modifier;
    }

    public void setUp() {
        super.setUp();
        Element topNode = null;
        if (jmethod != null) {
            topNode = jmethod.getElem();
        }
        if (jsinit != null) {
            topNode = jsinit.getElem();
        }
        if (topNode != null) {
            markSwitchLabel(topNode);
            markLocalInForStatement(topNode);
        }
    }

    private void markSwitchLabel(Element topElem) {
        NodeList nl = topElem.getElementsByTagName(JXConstants.LabelElem);
        for (int i = 0; i < nl.getLength(); i++) {
            Element e = (Element)nl.item(i);

            Node parent = e.getParentNode();
            if (parent.getNodeName().equals(JXConstants.StatementElem)) {  // BLOCK
                parent = parent.getParentNode();

                JavaStatement jst = new JavaStatement(parent);
                if (jst.isSwitchStatement()) {
                    e.setAttribute(JXConstants.SortAttr, JXConstants.Switch);
                }
            }
        }
    }

    private void markLocalInForStatement(Element topElem) {
        NodeList nl = topElem.getElementsByTagName(JXConstants.LocalElem);
        for (int i = 0; i < nl.getLength(); i++) {
            Element e = (Element)nl.item(i);

            Node node = e;
            while (node != null && !node.getNodeName().equals(JXConstants.OperatorElem)) {
                node = node.getPreviousSibling();
            }
            
            if (node != null && node.getFirstChild().getNodeValue().equals("(")) {
                while (node != null && !node.getNodeName().equals(JXConstants.KeywordElem)) {
                    node = node.getPreviousSibling();
                }

                if (node != null && node.getFirstChild().getNodeValue().equals("for")) {
                    e.setAttribute(JXConstants.SortAttr, JXConstants.For);
                }
            }
        }
    }

    public void preconditions() throws RefactoringException {
        collectSelectedValidElements();
        
        containsValidElement();
        checkSelectedRange();
        checkSelectedCode();
        checkLocalStatement();
        checkLabelStatement();
        checkConstructorCall();

        try {
            if (jmethod != null) {
                cfg = new CFG(jmethod.getElem());
            } else {
                cfg = new CFG(jsinit.getElem());
            }
            pdg = new PDG(cfg);
            collectCFGPDGNodes();

        } catch (CFGException e) {
            throw new RefactoringException(e.getMessage());
        }

        checkLocalType();
        checkSelectedCodeByUsingCFGPDG();
        String returnType = getReturnVariableType();

        if (newName == null) {
            NewMethodNameDialog dialog = new NewMethodNameDialog(frame, "Extract Method");
            newName = dialog.getNewName();

            if (newName == null) {
                throw new RefactoringException("");
            }
            if (dialog.getPublicOption()) {
                modifier = Modifier.PUBLIC;
            } else if (dialog.getProtectedOption()) {
                modifier = Modifier.PROTECTED;
            } else if (dialog.getPrivateOption()) {
                modifier = Modifier.PRIVATE;
            } else {
                modifier = "";
            }
        }

        mcode = new MethodCode(returnType, newName);
        mcode.addModifier(modifier);
        checkAlreadyUsed(mcode);

        collectParameters(mcode);
    }

    public void transform() throws RefactoringException {
        transformer = new ExtractMethodTransformer(selectedValidElems, innerCFGNodes, outerCFGNodes,
                                                   jmethod, jsinit, cfg, mcode, rjv);
        transformer.changeSrc();
        unparse(jfile);
    }

    public void additionalTransformation() throws RefactoringException {
    }

    public String getLog() {
        if (jmethod != null) {
            return mcode.getName() + " from " + jmethod.getName();
        } else {
            return mcode.getName() + " from static initializer";
        }
    }

    private void collectSelectedValidElements() {
        Iterator it = selectedElems.iterator();
        while(it.hasNext()) {
            Element e = (Element)it.next();

            collectSelectedValidElements(e);
        }
    }

    private void collectSelectedValidElements(Node node) {
        while (node != null) {
            if (isValidElement(node)) {
                break;
            }
            node = node.getParentNode();
        }
    }

    private boolean isValidElement(Node node) {
        String nodeName = node.getNodeName();

        if (nodeName.equals(JXConstants.StatementElem) ||
            nodeName.equals(JXConstants.ClassElem) ||
            nodeName.equals(JXConstants.MethodElem) ||
            nodeName.equals(JXConstants.SpaceElem) ||
            nodeName.equals(JXConstants.NewLineElem) ||
            nodeName.equals(JXConstants.CommentElem) ||
            isIndependentLocalStatement(node) ||
            isIndependentLabelStatement(node) ||
            isSpecialSeparator(node)) {

            addSelectedValidElems(node);
            return true;
        }

        if (nodeName.equals(JXConstants.ExpressionElem) ||
            nodeName.equals(JXConstants.LocalElem) ||
            nodeName.equals(JXConstants.LabelElem) ||
            nodeName.equals(JXConstants.KeywordElem) ||
            nodeName.equals(JXConstants.OperatorElem)) {

            addSelectedValidElems(node);
            return false;

        }
        return false;
    }

    private void addSelectedValidElems(Node node) {
        if (!selectedValidElems.contains(node)) {
            selectedValidElems.add(node);
        }
    }

    private boolean isIndependentLocalStatement(Node node) {  // not part of the "for" statement
        if (node.getNodeName().equals(JXConstants.LocalElem)) {
            Node parent = node.getParentNode();
            return ((Element)node).getAttribute(JXConstants.SortAttr).length() == 0;
        }
        return false;
    }

    private boolean isIndependentLabelStatement(Node node) {  // not switch lable
        if (node.getNodeName().equals(JXConstants.LabelElem)) {
            return ((Element)node).getAttribute(JXConstants.SortAttr).length() == 0;
        }
        return false;
    }

    private boolean isSpecialSeparator(Node node) {
        if (node.getNodeName().equals(JXConstants.OperatorElem)) {
            String op = node.getFirstChild().getNodeValue();
            if (op.equals(";") || op.equals(",")) {
                return true;
            }
        }
        return false;
    }

    private void containsValidElement() throws RefactoringException {
        Iterator it = selectedValidElems.iterator();
        while (it.hasNext()) {
            Element e = (Element)it.next();

            if (e.getNodeName().equals(JXConstants.StatementElem) ||
                e.getNodeName().equals(JXConstants.LocalElem) ||
                e.getNodeName().equals(JXConstants.ClassElem) ||
                e.getNodeName().equals(JXConstants.InterfaceElem)) {
                return;
            }
        }
        
        throw new RefactoringException("Not extract:" +
          " An selection must at least one statement");
    }

    private void checkSelectedRange() throws RefactoringException {
        Iterator it = selectedValidElems.iterator();
        while (it.hasNext()) {
            Element e = (Element)it.next();

            if (!isEnclosedIn(e)) {
                throw new RefactoringException("Not extract:" +
                  " An expression (part of a statement) is not permitted to extract");
            }
        }
    }

    private boolean isEnclosedIn(Element e) {
        if ((e.getNodeName().equals(JXConstants.IdentElem) ||
             e.getNodeName().equals(JXConstants.OperatorElem) ||
             e.getNodeName().equals(JXConstants.KeywordElem) ||
             e.getNodeName().equals(JXConstants.LiteralElem)) && 
            !selectedElems.contains(e)) {
            return false;
        }

        if (!isEnclosedIn(e.getElementsByTagName(JXConstants.IdentElem)) ||
            !isEnclosedIn(e.getElementsByTagName(JXConstants.OperatorElem)) ||
            !isEnclosedIn(e.getElementsByTagName(JXConstants.KeywordElem)) ||
            !isEnclosedIn(e.getElementsByTagName(JXConstants.LiteralElem))) {
            return false;
        }
        return true;
    }

    private boolean isEnclosedIn(NodeList nl) {
        for (int i = 0; i < nl.getLength(); i++) {
            if (!selectedElems.contains((Element)nl.item(i))) {
                return false;
            }
        }
        return true;
    }
            
    private void checkSelectedCode() throws RefactoringException {
        if (elem == null) {
            throw new RefactoringException("Not extract:" +
              " the selection does not contains a statement");
        }

        if (jmethod == null && jsinit == null) {
            throw new RefactoringException("Not extract:" +
              " the selection is not enclosed by the single method or static initializer");
        }

        if (jmethod != null && !isEnclosedBySingleMethod()) {
            throw new RefactoringException("Not extract:" +
              " the selection is not enclosed by the single method");
        }

        if (jsinit != null && !isEnclosedBySingleStaticInitializer()) {
            throw new RefactoringException("Not extract:" +
              " the selection is not enclosed by the single static initializer");
        }
    }

    private boolean isEnclosedBySingleMethod() {
        Iterator it = selectedElems.iterator();
        while (it.hasNext()) {
            Element e = (Element)it.next();

            if (!isEnclosedBySingleMethod(e)) {
                return false;
            }
        }
        return true;
    }

    private boolean isEnclosedBySingleMethod(Node node) {
        JavaElement jelem = new JavaElement((Element)node);

        while (node != null) {
            JavaMethod jm = jelem.getJavaMethod(node);
            if (jmethod.equals(jm)) {
                return true;
            }
            node = node.getParentNode();
        }
        return false;
    }

    private boolean isEnclosedBySingleStaticInitializer() {
        Iterator it = selectedElems.iterator();
        while (it.hasNext()) {
            Element e = (Element)it.next();

            JavaElement jelem = new JavaElement(e);
            JavaStaticInit ji = jelem.getJavaStaticInit(e);
            if (!jsinit.equals(ji)) {
                return false;
            }
        }
        return true;
    }

    private void checkLocalStatement() throws RefactoringException {
        Iterator it = selectedValidElems.iterator();
        while (it.hasNext()) {
            Element e = (Element)it.next();
            
            if (e.getNodeName().equals(JXConstants.LocalElem)) {
                JavaLocal jl = new JavaLocal(e);

                Iterator it2 = jl.getTypeSharedLocalNodes().iterator();
                while (it2.hasNext()) {
                    Element e2 = (Element)it2.next();

                    if (!isEnclosedIn(e2)) {
                        throw new RefactoringException("Not extract:" +
                          " A part of a local declaration is not permitted to extract");
                    }
                }
            }
        }
    }

    private void checkLabelStatement() throws RefactoringException {
        Iterator it = selectedValidElems.iterator();
        while (it.hasNext()) {
            Element e = (Element)it.next();

            if (isIndependentLabelStatement(e)) {
                throw new RefactoringException("Not extract:" +
                  " the selection contains a label statement");
            }
        }
    }

    private void checkLocalType() throws RefactoringException {
        Iterator it = selectedValidElems.iterator();
        while (it.hasNext()) {
            Element e = (Element)it.next();

            if (e.getNodeName().equals(JXConstants.ClassElem) ||
                e.getNodeName().equals(JXConstants.InterfaceElem)) {
                
                boolean exists = false;
                Node parent = e.getParentNode();
                if (parent.getNodeName().equals(JXConstants.ExpressionElem)) {
                    JavaExpr jexpr = new JavaExpr(parent);
                    if (!jexpr.isInstanceCreation()) {
                        exists = true;
                    }
                } else {
                    exists = true;
                }

                JavaClass jc = new JavaClass(e);
                if (exists && usesLocalType(jc.getName(), outerCFGNodes)) {
                    throw new RefactoringException("Not extract:" +
                      " the selection contains a local type declaration referenced in the remaining code");
                }

            } else {
                NodeList nl = e.getElementsByTagName(JXConstants.TypeElem);
                for (int i = 0; i < nl.getLength(); i++) {
                    Element t = (Element)nl.item(i);

                    JavaType jt = new JavaType(t);
                    if (usesLocalClass(jt.getName(), outerCFGNodes)) {
                        throw new RefactoringException("Not extract:" +
                          " the selection contains a reference to a local type declared in the remaining code");
                    }
                }
            }
        }
    }

    private boolean usesLocalType(String className, GraphCompSet nodes) {
        Iterator it = nodes.iterator();
        while (it.hasNext()) {
            CFGNode cfgNode = (CFGNode)it.next();

            Element e = cfgNode.getElem();
            if (e != null) {
                NodeList nl = e.getElementsByTagName(JXConstants.TypeElem);
                for (int i = 0; i < nl.getLength(); i++) {
                    Element t = (Element)nl.item(i);

                    JavaType jt = new JavaType(t);
                    if (jt.getName().equals(className)) {
                        return true;
                    }
                }
            }
        }
        return false;
    }

    private boolean usesLocalClass(String className, GraphCompSet nodes) {
        Iterator it = nodes.iterator();
        while (it.hasNext()) {
            CFGNode cfgNode = (CFGNode)it.next();

            Element e = cfgNode.getElem();
            if (e != null) {
                NodeList nl = e.getElementsByTagName(JXConstants.ClassElem);
                for (int i = 0; i < nl.getLength(); i++) {
                    Element c = (Element)nl.item(i);

                    JavaClass jc = new JavaClass(c);
                    if (jc.getName().equals(className)) {
                        return true;
                    }
                }
            }
        }
        return false;
    }

    private void checkConstructorCall() throws RefactoringException {
        Iterator it = selectedValidElems.iterator();
        while (it.hasNext()) {
            Element e = (Element)it.next();

            if (e.getNodeName().equals(JXConstants.ExpressionElem)) {
                JavaExpr jexpr = new JavaExpr(e);

                if (jexpr.isConstructorCall() || jexpr.isSuperConstructorCall()) {
                    Element ek = (Element)jexpr.getFirstChildNode(e, JXConstants.KeywordElem);
                    if (ek.getNodeName().equals(JXConstants.KeywordElem)) {
                        String kw = ek.getFirstChild().getNodeValue();

                        if (kw.equals("this") || kw.equals("super")) {
                            throw new RefactoringException("Not extract:" +
                              " the selection contains a constructor call");
                        }
                    }
                }
            }
        }
    }

    private void checkSelectedCodeByUsingCFGPDG() throws RefactoringException {
        if (containReturn() && !isAllReturnFlow()) {
            throw new RefactoringException("Not extract:" +
              " the selection contains a return statement but an execution flow not ending in a return");
        }

        if (containsBranchWithUnselectedTarget()) {
            throw new RefactoringException("Not extract:" +
              " the selection contains a break or continue statement the target of which is not selected");
        }

        if (getEnclosingParentNodes().size() > 1) {
            throw new RefactoringException("Not extract:" +
              " the selection is not enclosed by the single parent statement");
        }
    }

    private void collectCFGPDGNodes() {
        innerCFGNodes = new GraphCompSet();
        innerPDGNodes = new GraphCompSet();

        Iterator it = selectedValidElems.iterator();
        while (it.hasNext()) {
            Element e = (Element)it.next();
            collectCFGPDGNodes(e, innerCFGNodes, innerPDGNodes);
        }
        adjustMergeNode();

        outerCFGNodes = cfg.getNodes().difference(innerCFGNodes);
    }

    private void collectCFGPDGNodes(Element e, GraphCompSet cfgNodes, GraphCompSet pdgNodes) {
        String nodeName = e.getNodeName();
        if (nodeName.equals(JXConstants.StatementElem) ||
            nodeName.equals(JXConstants.LocalElem) ||
            nodeName.equals(JXConstants.LabelElem)) {

            CFGNode cfgNode = cfg.getNode(e);
            if (cfgNode != null) {
                cfgNodes.add(cfgNode);
                pdgNodes.add(getPDGNode(cfgNode));
            }
            
            NodeList nl = e.getElementsByTagName(JXConstants.ExpressionElem);
            for (int i = 0; i < nl.getLength(); i++) {
                cfgNode = cfg.getNode((Element)nl.item(i));
                if (cfgNode != null) {
                    cfgNodes.add(cfgNode);
                    pdgNodes.add(getPDGNode(cfgNode));
                }
            }
        }
    }

    private PDGNode getPDGNode(CFGNode cfgNode) {
        Iterator it = pdg.getNodes().iterator();
        while (it.hasNext()) {
            PDGNode node = (PDGNode)it.next();

            if (node.getCFGNode().equals(cfgNode)) {
                return node;
            }
        }
        return null;
    }

    private void adjustMergeNode() {
        Iterator it = cfg.getNodes().iterator();
        while (it.hasNext()) {
            CFGNode cfgNode = (CFGNode)it.next();

            if (cfgNode.isMergeSt() && toBeInInnerNodes(cfgNode)) {
                innerCFGNodes.add(cfgNode);
            }
        }
    }

    private boolean toBeInInnerNodes(CFGNode cfgNode) {
        Iterator it = cfgNode.getSrcNodes().iterator();
        while (it.hasNext()) {
            CFGNode n = (CFGNode)it.next();

            if (!innerCFGNodes.contains(n)) {
                return false;
            }
        }
        return true;
    }

    private boolean containReturn() {
        Iterator it = selectedValidElems.iterator();
        while (it.hasNext()) {
            Element e = (Element)it.next();

            if (e.getNodeName().equals(JXConstants.StatementElem)) {
                JavaStatement jst = new JavaStatement(e);
                if (jst.isReturnStatement()) {
                    return true;
                }
            }
        }
        return false;
    }

    private boolean isAllReturnFlow() {
        //cfg.print();
        Iterator it = innerCFGNodes.iterator();
        while (it.hasNext()) {
            CFGNode cfgNode = (CFGNode)it.next();

            if (!cfgNode.isReturnSt() && existsOuterSuccessor(cfgNode)) {
                return false;
            }
        }
        return true;
    }

    private boolean existsOuterSuccessor(CFGNode cfgNode) {
        Iterator it = cfgNode.getSuccessors().iterator();
        while (it.hasNext()) {
            CFGNode node = (CFGNode)it.next();

            if (!node.isEntrySt() && !node.isExitSt() && !node.isFormalOutSt() &&
                outerCFGNodes.contains(node)) {
                return true;
            }
        }
        return false;
    }

    private boolean containsBranchWithUnselectedTarget() {
        //cfg.print();
        Iterator it = innerCFGNodes.iterator();
        while (it.hasNext()) {
            CFGNode cfgNode = (CFGNode)it.next();

            if (cfgNode.isBreakSt() || cfgNode.isContinueSt()) {

                CFGNode target = getDstNodeOfTrueFlow(cfgNode);
                if (target != null && !innerCFGNodes.contains(target)) {
                    return true;
                }
            }
        }
        return false;
    }

    private CFGNode getDstNodeOfTrueFlow(CFGNode node) {
        Iterator it = node.getOutgoingEdges().iterator();
        while (it.hasNext()) {
            Flow flow = (Flow)it.next();

            if (flow.isTrue()) {
                return (CFGNode)flow.getDstNode();
            }
        }
        return null;
    }

    private GraphCompSet getEnclosingParentNodes() {
        GraphCompSet parents = new GraphCompSet();

        Iterator it = pdg.getEdges().iterator();
        while (it.hasNext()) {
            Dependence dependence = (Dependence)it.next();
            if (dependence.isCD()) {
                PDGNode src = (PDGNode)dependence.getSrcNode();
                PDGNode dst = (PDGNode)dependence.getDstNode();
                if (!src.getCFGNode().isDoSt() &&
                    !innerPDGNodes.contains(src) && innerPDGNodes.contains(dst)) {
                    parents.add(src);
                }
            }
        }

        //cfg.print();
        //pdg.print();
        //System.out.println(parents.size());
        return parents;
    }
    
    private String getReturnVariableType() throws RefactoringException {
        JavaVariableList vars = getReturnJavaVariables();
        if (vars.size() > 1) {
            throw new RefactoringException("Not extract:" +
              " the selection contains more than one return variable");

        } else if (vars.size() == 1) {
            rjv = vars.getFirst();
            if (QueryClass.hasImported(rjv.getJavaClass(), rjv.getType())) {
                int dot = rjv.getType().lastIndexOf(".");
                if (dot != -1) {
                    return rjv.getType().substring(dot + 1);
                }
            }
            return rjv.getType();
        }
        return "void";
    }

    private JavaVariableList getReturnJavaVariables() {
        JavaVariableList vars = new JavaVariableList();

        //pdg.print();
        Iterator it = pdg.getEdges().iterator();
        while (it.hasNext()) {
            Dependence edge = (Dependence)it.next();

            if (edge.isDD()) {
                DD dd = (DD)edge;
                GraphNode src = dd.getSrcNode();
                GraphNode dst = dd.getDstNode();
                GraphNode lc = dd.getLoopCarriedNode();

                if (innerPDGNodes.contains(src) &&
                    (!innerPDGNodes.contains(dst) || (lc != null && !innerPDGNodes.contains(lc)))) {

                    JavaVariable jv =  dd.getVariable();
                    if ((jv.isLocal() || jv.isParameter() || jv.isMethodReturn()) && 
                        !contains(vars, jv)) {
                        vars.add(jv);
                    }
                }
            }
        }
        return vars;
    }

    private void checkAlreadyUsed(MethodCode mc) throws RefactoringException {
        if (QueryClass.hasTheMethod(jclass, mc.getSignature())) {
            throw new RefactoringException("Not rename:" +
              " the method " + mc.getName() + " already exists in the class " + jclass.getName());
        }
    }

    private void collectParameters(MethodCode code) {
        JavaVariableList vars = new JavaVariableList();

        //pdg.print();
        Iterator it = pdg.getEdges().iterator();
        while (it.hasNext()) {
            Dependence edge = (Dependence)it.next();

            if (edge.isDD()) {
                DD dd = (DD)edge;
                GraphNode src = dd.getSrcNode();
                GraphNode dst = dd.getDstNode();

                if (!innerPDGNodes.contains(src) && innerPDGNodes.contains(dst)) {
                    JavaVariable jv =  dd.getVariable();
                    if ((jv.isLocal() || jv.isParameter()) && !contains(vars, jv)) {
                        vars.add(jv);
                        JavaLocal jl = new JavaLocal(jv.getDeclaration());
                        if (jl.isFinal()) {
                            code.addParameter(jv.getType(), jv.getTypeFqn(), jv.getName(), true);
                        } else {
                            code.addParameter(jv.getType(), jv.getTypeFqn(), jv.getName());
                        }
                    }
                }
            }
        }
    }

    private boolean contains(JavaVariableList vars, JavaVariable jvar) {
        Iterator it = vars.iterator();
        while (it.hasNext()) {
            JavaVariable jv = (JavaVariable)it.next();

            if (jv.getName().equals(jvar.getName())) {
                return true;
            }
        }
        return false;
    }
}
