/*
 *     ExtractMethodTransformer.java  Aug 29, 2004
 *
 *     Katsuhisa Maruyama (maru@jtool.org)
 */

package org.jtool.jrbx.refactor.miscellaneous;
import org.jtool.jrbx.refactor.*;
import org.jtool.jrbx.refactor.util.*;
import org.jtool.jrbx.refactor.query.*;
import org.jtool.platform.xml.JXConstants;
import org.jtool.platform.xml.wrapper.*;
import org.jtool.platform.xml.graph.*;
import org.jtool.platform.xml.graph.cfg.*;
import org.jtool.platform.xml.graph.pdg.*;
import java.util.*;
import org.w3c.dom.*;

public class ExtractMethodTransformer extends RefactoringTransformer {
    private ArrayList selectedValidElems;
    private CFG cfg;
    private MethodCode mcode;
    private JavaVariable rjv;
    private JavaMethod jmethod;
    private JavaStaticInit jsinit;
    private GraphCompSet innerCFGNodes;
    private GraphCompSet outerCFGNodes;

    public ExtractMethodTransformer(ArrayList elems, GraphCompSet innerNodes, GraphCompSet outerNodes,
                                    JavaMethod jmethod, JavaStaticInit jsinit,
                                    CFG cfg, MethodCode mcode, JavaVariable rjv) {
        selectedValidElems = elems;
        innerCFGNodes = innerNodes;
        outerCFGNodes = outerNodes;

        this.jmethod = jmethod;
        this.jsinit = jsinit;
        this.cfg = cfg;
        this.mcode = mcode;
        this.rjv = rjv;
    }

    public void changeSrc() {
        insertDeclarationInExtractedMethod();
        insertDeclarationInRemainingMethod();

        String code = createExtractedMethodCode();
        if (jmethod != null) {
            insertCodeAfter("\n\n" + code, jmethod.getElem());
        } else {
            insertCodeAfter("\n\n" + code, jsinit.getElem());
        }
        deleteExtractedMethodCode();
        insertMethodCall();
    }

    private void insertMethodCall() {
        StringBuffer buf = new StringBuffer();

        if (rjv != null) {
            if (rjv.isMethodReturn()) {
                buf.append("return ");

            } else {
                if (!isDeclared(rjv)) {
                    buf.append(rjv.getType());
                    buf.append(" ");
                }
                buf.append(rjv.getName());
                buf.append(" = ");
            }
        }

        buf.append(mcode.getName());
        buf.append("(");
        buf.append(mcode.getParameterNames());
        buf.append(");");

        insertCodeBefore(buf.toString(), (Element)selectedValidElems.get(0));
    }

    private String createExtractedMethodCode() {
        if ((jmethod != null && jmethod.isStatic()) || 
            jsinit != null && jsinit.isStatic()) {
            mcode.addModifier(Modifier.STATIC);
        }

        getThrows();

        StringBuffer buf = new StringBuffer();
        buf.append("    ");
        buf.append(mcode.getCode());
        buf.append(getThrows());
        buf.append(" {\n");
        buf.append(getContents());
        
        if (rjv != null && !rjv.isMethodReturn()) {
            buf.append("\n        ");
            buf.append("return ");
            buf.append(rjv.getName());
            buf.append(";\n");
        }
        buf.append("    }");

        return buf.toString();
    }
    
    private String getThrows() {
        ArrayList receivers = new ArrayList();

        Iterator it = innerCFGNodes.iterator();
        while (it.hasNext()) {
            CFGNode cfgNode = (CFGNode)it.next();

            if (cfgNode.isCallSt()) {
                JavaCall ja = new JavaCall(cfgNode.getElem());

                JavaMethod jm = ja.getCalledJavaMethod();
                if (jm != null) {
                    receivers.addAll(jm.getThrowsNodes());
                }
            }
        }

        StringBuffer buf = new StringBuffer();
        for (int i = 0; i < receivers.size(); i++) {
            if (i == 0) {
                buf.append(" throws ");
            } else {
                buf.append(" ,");
            }
            Element elem = (Element)receivers.get(i);
            JavaElement je = new JavaElement(elem);
            buf.append(je.getElemType(elem));
        }

        if (receivers.size() == 0) {
            return "";
        } else {
            return buf.toString();
        }
    }

    private String getContents() {
        StringBuffer buf = new StringBuffer();

        Iterator it = selectedValidElems.iterator();
        while(it.hasNext()) {
            Element elem = (Element)it.next();

            if (!containsParents(selectedValidElems, elem.getParentNode())) {
                String code = getContentsOfCode(elem);
                buf.append(code);

                JavaElement jelem = new JavaElement(elem);
                if (jelem.isFieldNode(elem) || jelem.isLocalNode(elem) ||
                    jelem.isClassNode(elem) || jelem.isMethodNode(elem) ||
                    jelem.isStatementNode(elem) || jelem.isExpressionNode(elem)) {

                } else if (jelem.isOperatorNode(elem)) {
                    if (code.equals("{") || code.equals("}") || code.equals(";")) {
                        newLine(buf, elem);
                    } else if (code.equals(",")) {
                        buf.append(" ");
                    }

                } else {
                    newLine(buf, elem);
                }
            }
        }
        return buf.toString();
    }

    private void newLine(StringBuffer buf, Node node) {
        if (node.getNodeName().equals(JXConstants.NewLineElem)) {
            return;
        }

        node = node.getPreviousSibling();
        while (node != null && node.getNodeName().equals(JXConstants.SpaceElem)) {
            node = node.getPreviousSibling();
        }
        if (node != null && !node.getNodeName().equals(JXConstants.NewLineElem)) {
            buf.append("\n");
        }
    }        
    
    private boolean containsParents(ArrayList nodes, Node node) {
        while (node.getNodeType() == Node.ELEMENT_NODE && node != null) {
            if (contains(nodes, (Element)node)) {
                return true;
            }
            node = node.getParentNode();
        }
        return false;
    }

    private boolean contains(ArrayList nodes, Element elem) {
        Iterator it = nodes.iterator();
        while (it.hasNext()) {
            Element e = (Element)it.next();

            if (e.equals(elem)) {
                return true;
            }
        }
        return false;
    }

    private boolean isDeclared(JavaVariable jv) {
        Element e = (Element)jv.getDeclaration();
        CFGNode declNode = cfg.getNode(e);
        return outerCFGNodes.contains(declNode);
    }

    private void deleteExtractedMethodCode() {
        Iterator it = selectedValidElems.iterator();
        while (it.hasNext()) {
            Element elem = (Element)it.next();

            JavaElement jelem = new JavaElement(elem);

            if (jelem.isFieldNode(elem)) {
                delete(new JavaField(elem));

            } else if (jelem.isLocalNode(elem)) {
                delete(new JavaLocal(elem));

            } else if (jelem.isExpressionNode(elem) && isStatementExpression(elem)) {
                delete(elem.getParentNode());

            } else {
                delete(elem);
            }
        }
    }

    private boolean isStatementExpression(Element elem) {
        Element pelem = (Element)elem.getParentNode();
        JavaElement jelem = new JavaElement(pelem);
        if (jelem.isStatementNode(pelem)) {

            JavaStatement jst = new JavaStatement(pelem);
            if (jst.isExpressionStatement()) {
                return true;
            }
        }
        return false;
    }

    private void insertDeclarationInExtractedMethod() {
        Iterator it = getVariablesNeededToDeclaredInExtractedMethod().iterator();
        while (it.hasNext()) {
            JavaVariable jv = (JavaVariable)it.next();

            CFGNode newDeclNode = QueryGraph.getDeclaraionCFGNode(cfg, innerCFGNodes, jv);
            insertDeclaration(newDeclNode, jv,  innerCFGNodes);
        }
    }

    private JavaVariableList getVariablesNeededToDeclaredInExtractedMethod() {
        JavaVariableList vars = new JavaVariableList();

        Iterator it = innerCFGNodes.iterator();
        while(it.hasNext()) {
            CFGNode cfgNode = (CFGNode)it.next();
            if (cfgNode.hasDefVariable() && !cfgNode.isLocalSt()) {
                CFGStatement node = (CFGStatement)cfgNode;

                JavaVariable jv = node.getDefVars().getFirst();
                if (jv != null && !hasDeclaredAsParameter(jv.getName())) {

                    Element e = (Element)jv.getDeclaration();
                    if (e != null) {
                        CFGNode declNode = cfg.getNode(e);
                        if (declNode != null &&
                            !innerCFGNodes.contains(declNode) && !contains(vars, jv)) {
                            vars.add(jv);
                        }
                    }
                }
            }
        }
        return vars;
    }

    private boolean hasDeclaredAsParameter(String name) {
        for (int i = 0; i < mcode.getNumOfParameter(); i++) {
            String param = mcode.getParameterName(i);
            if (param.equals(name)) {
                return true;
            }
        }
        return false;
    }

    private void insertDeclarationInRemainingMethod() {
        Iterator it = getVariablesNeededToDeclaredInRemainingMethod().iterator();
        while (it.hasNext()) {
            JavaVariable jv = (JavaVariable)it.next();

            CFGNode newDeclNode = QueryGraph.getDeclaraionCFGNode(cfg, outerCFGNodes, jv);
            insertDeclaration(newDeclNode, jv, outerCFGNodes);
        }
    }

    private JavaVariableList getVariablesNeededToDeclaredInRemainingMethod() {
        JavaVariableList vars = new JavaVariableList();

        Iterator it = outerCFGNodes.iterator();
        while(it.hasNext()) {
            CFGNode cfgNode = (CFGNode)it.next();
            if (cfgNode.hasDefVariable() && !cfgNode.isLocalSt()) {
                CFGStatement node = (CFGStatement)cfgNode;

                JavaVariable jv = node.getDefVars().getFirst();
                if (jv != null && !jv.equals(rjv)) {

                    Element e = (Element)jv.getDeclaration();
                    if (e != null) {
                        CFGNode declNode = cfg.getNode(e);
                        if (declNode != null &&
                            !outerCFGNodes.contains(declNode) && !contains(vars, jv)) {
                            vars.add(jv);
                        }
                    }
                }
            }
        }
        return vars;
    }

    private boolean contains(JavaVariableList vars, JavaVariable jv) {
        Node node = jv.getDeclaration();
        Iterator it = vars.iterator();
        while (it.hasNext()) {
            JavaVariable v = (JavaVariable)it.next();
            if (v.getDeclaration().equals(jv.getDeclaration())) {
                return true;
            }
        }
        return false;
    }

    private void insertDeclaration(CFGNode declNode, JavaVariable jv, GraphCompSet nodes) {
        if (declNode != null) {

            if (!nodes.contains(declNode)) {
                Element elem = (Element)selectedValidElems.get(0);
                insertDeclaration(elem, jv);

            } else if (declNode.isLocalSt()) {
                // nothing to do

            } else if (hasDefVariable(declNode, jv)) {
                insertCodeBefore(jv.getType() + " ", jv.getElem());

            } else {
                insertDeclaration(declNode.getElem(), jv);
            }
        }
    }

    private void insertDeclaration(Element elem, JavaVariable jv) {
        String declCode = jv.getType() + " " + jv.getName() + ";\n" + getSpace(elem);
        insertCodeBefore(declCode, elem);
    }

    private boolean hasDefVariable(CFGNode cfgNode, JavaVariable jv) {
        if (cfgNode.hasDefVariable()) {
            CFGStatement node = (CFGStatement)cfgNode;
            if (node.containsDefVariable(jv)) {
                return true;
            }
        }
        return false;
    }
}
